// @ myth - client - screen.h
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ begin

#ifndef SCREEN_HEADER
#define SCREEN_HEADER

// @@ include

#include "common/common.h"
#include "common/buffer.h"
#include "common/picture.h"
#include <sgtty.h>

// @@ extern

extern pfVpcC set_screen_handler(pfVpcC screen_handler);
extern void set_output_char(pfVC output_char);
extern void tputs(const char* cp, int affcnt = 1);

// @@ class

// @@@ TTY

class TTY : public BUFFER
{
    ltchars Old_ltchars; 
    sgttyb Old_sgttyb;
    tchars Old_tchars;
    char tmp;			// Used by put and get
public:
    inline void put(char c)
    { BUFFER::put(&c, sizeof(char)); }
    inline char get()
    { BUFFER::get(&tmp, sizeof(char)); return tmp; }
    inline bool good()
    { return GoodInput() && GoodOutput(); }
    inline char EraseChar()
    { return Old_sgttyb.sg_erase; }
    inline char EraseLine()
    { return Old_sgttyb.sg_kill; }
    inline char EraseWord()
    { return Old_ltchars.t_werasc; }
    inline char RefreshChar()
    { return Old_ltchars.t_rprntc; }
    inline char SuspendChar()
    { return Old_ltchars.t_suspc; }
    inline char QuitChar()
    { return Old_tchars.t_intrc; }
    TTY(const char* name);
    ~TTY();
};

// @@@ TERMCAP

class TERMCAP
{
    char buffer[1024];
    int TC_columns;		// co - number of columns on each line
    int TC_lines;		// li - number of lines in each column
    char* TC_left;		// bc - move cursor left (bs - ^H)
    char* TC_goto;		// cm - move cursor to position
    char* TC_clear;		// cl - clear screen
    char* TC_line;		// ce - clear to end of line
    char* TC_beep;		// bl - make a beep
    char* TC_region;		// cs - set scroll region
    char* TC_endVisual;		// ve - end visual mode
    char* TC_endCursor;		// te - end cursor addressing mode
public:
    inline coor size()
    { return coor(TC_columns, TC_lines); }
    inline bool CanClearLine()
    { return TC_line != NULL; }
    inline bool CanScrollRegion()
    { return TC_region != NULL; }
    inline void cursor(coor pos)
    { tputs(tgoto(TC_goto, pos.x, pos.y)); }
    inline void BackSpace()
    { tputs(TC_left); }
    inline void ClearScreen()
    { tputs(TC_clear); }
    inline void ClearLine()
    { tputs(TC_line); }
    inline void SetScrollRegion(int first, int last)
    { tputs(tgoto(TC_region, first, last)); }
    inline void beep()
    { tputs(TC_beep); }
    TERMCAP();
    ~TERMCAP();
};

// @@@ SCREEN

class SCREEN : TERMCAP
{
    PICTURE picture;		// Current content of screen (without window)
    coor cpos;			// The  cursors position on screen
    PICTURE window;		// Overlaping Window
    coor wpos;			// UpperRight corner of window
    int RegionFirst;		// Scroll Region
    int RegionLast;
    inline bool OnEdge(coor pos)
    { return  pos.x == size().x - 1; }
    inline bool OnCorner(coor pos)
    { return pos == size() - coor(1, 1); }
    inline bool OnWindow(coor pos, const PICTURE& window, coor wpos)
    { return pos >= wpos && pos < wpos + window.size(); }
    inline bool OnWindow(coor pos)
    { return OnWindow(pos, window, wpos); }
    inline void SetCursor(coor pos)
    { cpos = pos; TERMCAP::cursor(cpos); }
    inline void LoseCursor()
    { cpos = NoWhere; }
    char screen(coor pos, const PICTURE& window, coor wpos);
    inline char screen(coor pos)
    { return screen(pos, window, wpos); }
    bool IsPictureBlank(const PICTURE& picture, coor from, coor to);
    bool IsScreenBlank(coor from, coor to);
public:
    inline bool OnScreen(coor pos)
    { return pos >= Origo && pos < size(); }
    inline char what(coor pos)
    { return OnScreen(pos) ? picture[pos] : ' '; }
    void plot(coor pos, char c);
    void print(coor pos, const char* s);
    void ClearLine(coor from, coor to);
    void cursor(coor pos);
    void ClearScreen();
    void ScrollRegion(int first, int last);
    void redraw();
    void cover(const PICTURE& window, coor pos);
    void cover();
    inline void cover(const PICTURE& window)
    { cover(window, wpos); }
    inline coor size()
    { return TERMCAP::size(); }
    inline void beep()
    { TERMCAP::beep(); }
    SCREEN();
};

// @@ end

#endif SCREEN_HEADER

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
