// @ myth - common - ipc.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk
//
// These rutines implement machine independent interprocess
// communication facilities between a server process and one or more
// client processes.
//
// The server creates a GATE by defining one:
//    GATE gate
//
// The client connect to the GATE by defining a CHANNEL:
//    CHANNEL server
//
// The server can accept this by defining a CHANNEL itself:
//    CHANNEL client(&gate)
//
// Now the client can write to server with:
//    server.PutChar('a')
//    server.PutShort(10)
//    server.PutLong(1000000L)
//    server.PutString("Hello\n")
//    server.PutCoor(coor(10,5))
//
// The server can read from the client with:
//    char c = client.GetChar()
//    short i = client.GetShort()
//    long l = client.GetLong()
//    STRING s = client.GetString()
//    coor pos = client.GetCoor()
//
// The channel is buffered. The buffer will be emptied when the
// channel is deleted, or by a call to
//    server.flush()
//
// The channel is bidirectional, so it is possible to do something like:
//    server.PutString("Hello\n")
//    long l = server.GetLong()
//
// All shorts are truncated to the length of 15 bit + sign, and all longs
// are truncated to 31 bit + sign. This is done in order to make the
// transfer machine independent.
//
// It is the responsibility of the user to make sure date is read and
// writen the same way, eg NOT like this:
//    server.PutShort(10)
//    // ...
//    char garbage = client.GetChar() // Will read half a short...
//
// If there is any problems with writing or reading, the following
// will be FALSE:
//    if(client.GoodInput() && client.GoodOutput()) // No problems...
//
// Created by:
//    Per Abrahamsen alias amanda@iesd.uucp
//

// @@ include

#include "common/ipc.h"
#include "common/error.h"
#include "common/parse.h"
#include "common/format.h"
#include <errno.h>

// @@ static err step(int tries)
//
// The default rutine used when waiting for a connection.

static err step(int tries)
{
    sleep(1);
    (void) tries;
    return SUCCES;
}

// @@ const

// Some arbitrary size of the buffer. Any reason to prefer some special size?
static const int ChannelBufferSize = 4096;

// Filename containing active host
static const char* const HostFile = "host";

// @@ GATE

// @@@ GATE::GATE()

GATE::GATE() : (BUFFER_NONE)
{
    // Log new server's name in HostFile

    BUFFER LogFile(BUFFER_EXCLUSIVE, format("%s/%s/%s",
					    HomeDirectory,
					    LibraryDirectory, HostFile));

    if(LogFile.fileno() == ERROR)
	// Not really an error' but what else to do?
	bug.panic("GATE::GATE", "Couldn't open LogFile");

    char HostName[MessageSize];
    if(gethostname(HostName, MessageSize - 1) == ERROR)
	bug.panic("GATE::GATE", "::gethostname() failed");
    (void) strcat(HostName, "\n");

    LogFile.put(HostName, strlen(HostName));

    // Create socket

    sockaddr_in address;
    address.sin_family = AF_INET;
    address.sin_port = InternetPort;
    address.sin_addr.s_addr = INADDR_ANY;
    
    int fd = socket(AF_INET, SOCK_STREAM, 0);
    
    if(fd < 0 )
	bug.panic("GATE::GATE", "socket() fail");
    
    if(bind(fd, &address, sizeof(address)) == ERROR)
	bug.panic("GATE::GATE", "bind() fail");
    
    if(listen(fd, SOMAXCONN) == ERROR)
	bug.panic("GATE::GATE", "listen() fail");

    open(fd);
}

// @@@ GATE::~GATE()

GATE::~GATE()
{
    if(fileno() == ERROR)
	return;

    if(shutdown(fileno(), 2) == ERROR)
	bug.error("GATE::~GATE", "::shutdown() fail");

    if(unlink(format("%s/%s/%s", HomeDirectory,
		     LibraryDirectory, HostFile).string()) == ERROR)
	bug.error("GATE::~GATE", "::unlink() failed");
}

// @@ CHANNEL

// @@@ void CHANNEL::fail(const char* where, const char* what)

void CHANNEL::fail(const char* where, const char* what)
{
    GotError = TRUE;
    bug.warning(where, what);
}

// @@@ CHANNEL::CHANNEL(const char* name, err (*step)(int tries) = NULL)
//
// Try to open a channel to the server hosted on `name'.
// Call <step> between each trie.

CHANNEL::CHANNEL(const char* name, pfei step = NULL) : (BUFFER_RDWR)
{
    // No error detected.
    GotError = FALSE;
    
    if(step == NULL)
	step = ::step;

    hostent* host = gethostbyname(name);

    if(host == NULL)
    {
	fail("CHANNEL::CHANNEL", "Don't know host");
	return;
    }
    sockaddr_in address;	// socket address
    address.sin_family = AF_INET;
    address.sin_port = InternetPort;
    bcopy(host->h_addr, &address.sin_addr, host->h_length);

    int fd = socket(AF_INET, SOCK_STREAM, 0);
    if(fd == ERROR)
	bug.panic("CHANNEL::CHANNEL", "socket() failed");

    int tries = 1;		// Number of tries for connection.
    
    while(connect(fd, (struct sockaddr*) &address, sizeof(address)) == ERROR)
    {
	if(errno != ENOENT && errno != ECONNREFUSED)
	{
	    fail("CHANNEL::CHANNEL", "connect() failed");
	    (void) close(fd);
	    return;
	}
	else if(step(tries++) == ERROR)
	{
	    fail("CHANNEL::CHANNEL", "step() failed");
	    (void) close(fd);
	    return;
	}
	else
	    bug.warning("CHANNEL::CHANNEL", "connect() failed");
    }
    open(fd);
}

// @@@ CHANNEL::CHANNEL(const GATE* gate)

CHANNEL::CHANNEL(const GATE* gate) : (BUFFER_RDWR)
{
    GotError = FALSE;
    
    sockaddr_in address;	// socket address
    int length = sizeof(address); // socket address length

    int fd = accept(gate->fileno(), (struct sockaddr*) &address, &length);
    if(fd == ERROR)
    {
	fail("CHANNEL::CHANNEL", "accept() failed");
	return;
    }
    open(fd);
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
