// @ myth - server - cave.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "server/level.h"
#include "common/error.h"
#include "common/parse.h"
#include "server/rnd.h"
#include "server/stair.h"
#include "server/player.h"
#include "common/format.h"

// @@ LEVEL

// @@@ char LEVEL::ParseValue(PARSE* parse)

// @@@@ static const int LIT

static const int LIT = -2;

// @@@@ static const struct RockTable[]

static const struct { const char* string; int value; } RockTable[] =
{
    "FLOOR", FLOOR,
    "UNEVEN_FLOOR", UNEVEN_FLOOR,
    "VERY_SOFT_ROCK", VERY_SOFT_ROCK,
    "SOFT_ROCK", SOFT_ROCK,
    "ROCK", ROCK,
    "HARD_ROCK", HARD_ROCK,
    "VERY_HARD_ROCK", VERY_HARD_ROCK,
    "PERMANENT_ROCK", PERMANENT_ROCK,
    "LIT", LIT,
    NULL, ROCK_ERROR
};

// @@@@ static int RockToType(const char* string)

static int RockToType(const char* string)
{
    for(int i = 0; RockTable[i].string != NULL; i++)
    {
	if(strcmp(string, RockTable[i].string) == EQUAL)
	    return RockTable[i].value;
    }
    return ROCK_ERROR;
}

// @@@@ char LEVEL::ParseValue(PARSE* parse)

char LEVEL::ParseValue(PARSE* parse)
{
    int value = MAP_MASK_BIT;		// A good old PASCAL return value

    while(parse->good())
    {
	if(parse->LookAhead(INPUT_SPACE))
	    parse->skip(INPUT_SPACE);

	if(parse->LookAhead(INPUT_NEWLINE))
	    break;

	const char* word = parse->GetWord();
	int rock = RockToType(word);

	switch(rock)
	{
	case LIT:
	    value |= MAP_MASK_LIT;
	    break;
	case ROCK_ERROR:
	    parse->wrong("LEVEL:ParseValue", format("Unknown rock %s", word));
	    break;
	default:
	    value |= (rock << MAP_OFFSET_ROCK);
	    break;
	}
    }
    return value;
}

// @@@ void LEVEL::LoadText(PARSE* parse)

// @@@@ enum order_type

enum order_type
{
    ORDER_SYMBOL,		// Select symbols used to display rock
    ORDER_DEFAULT,		// Select default kind of rock
    ORDER_MAP,			// Read a map description
    ORDER_BUNCH,		// Read a bunch description
    ORDER_EXPERIENCE,		// Select experience level
    ORDER_SAVE,			// Promille chance to save the binary
    ORDER_CREATE,		// Create a random cave
    ORDER_POPULATION,		// The population in the cave
    ORDER_END,
    ORDER_ERROR = ERROR
};

// @@@@ static const struct OrderTable[]

static const struct { const char* string; order_type value; } OrderTable[] =
{
    "SYMBOL", ORDER_SYMBOL,
    "DEFAULT", ORDER_DEFAULT,
    "MAP", ORDER_MAP,
    "BUNCH", ORDER_BUNCH,
    "EXPERIENCE", ORDER_EXPERIENCE,
    "SAVE", ORDER_SAVE,
    "CREATE", ORDER_CREATE,
    "POPULATION", ORDER_POPULATION,
    NULL, ORDER_ERROR
};

// @@@@ static order_type OrderToType(const char* string)

static order_type OrderToType(const char* string)
{
    

    for(int i = 0; OrderTable[i].string != NULL; i++)
    {
	if(strcmp(string, OrderTable[i].string) == EQUAL)
	    return OrderTable[i].value;
    }
    return ORDER_ERROR;
}

// @@@@ void LEVEL::LoadText(PARSE* parse)

void LEVEL::LoadText(PARSE* parse)
{
    int value[MaxChar];		// Rock and lit represented by each char
    const char* order;		// Command order from parse file

    for(int i = 0; i < MaxChar; i++)
	value[i] = ERROR;

    if(parse->LookAhead(INPUT_COMMENT))
	parse->skip(INPUT_COMMENT);

    while(parse->good())
    {
	if(parse->LookAhead(INPUT_KEY)) // Location names
	{
	    parse->skip("\"");
	    char key = parse->GetKey();
	    parse->skip("\"");
	    value[key] = ParseValue(parse);
	}
	else switch(OrderToType(order = parse->GetWord()))
	{
	case ORDER_SYMBOL:
	    parse->skip(INPUT_SPACE);
	    strncpy(RockSymbol, parse->GetString(), ROCK_END);
	    break;
	case ORDER_DEFAULT:
	    map.SelectDefault(ParseValue(parse));
	    break;
	case ORDER_MAP:
	{
	    parse->skip(INPUT_SPACE);
	    /* const */ int MapY = (int) parse->GetLong();
	    parse->skip(INPUT_SPACE);
	    /* const */ int MapX = (int) parse->GetLong();
	    parse->skip(INPUT_COMMENT);
	    for(coor pos(0, MapY);
		parse->good() && parse->LookAhead(">");
		pos.y++)
	    {
		parse->skip(">");
		for(pos.x = MapX;
		    parse->good() && !parse->LookAhead(INPUT_NEWLINE);
		    pos.x++)
		{
		    char c = parse->GetKey();
		    if(value[c] == ERROR)
			parse->wrong("LEVEL::LoadText",
				     format("No value %c", c));

		    else
		    {
			map.plot(pos, value[c]);

			if(IsRock(rock(pos)))
			    /* Do Nothing */;
			else if(frequence.MonsterChance())
			{
			    MONSTER* monster =
				frequence.CreateMonster(experience());
			    AddMonster(monster);
			    monster->NewLevel(depth(), SYMBOL_ERROR);
			    monster->move(this, pos);
			}
			else if(frequence.ObjectChance())
			{
			    OBJECT* object = CreateThing(experience());
			    BUNCH* bunch = create(pos);
			    bunch->add(object);
			}
		    }
		    location[c] = pos;
		}
		parse->skip(INPUT_COMMENT);
	    }
	    break;
	}
	case ORDER_BUNCH:
	{
	    parse->skip(INPUT_SPACE);
	    parse->skip("\"");
	    coor where = locate(parse->GetKey());
	    parse->skip("\"");
	    parse->skip(INPUT_COMMENT);

	    // Insert objects here
	    OBJECT_LIST list;
	    list.load(parse, where);
	    BUNCH* bunch = create(where);
	    bunch->add(&list);

	    // Insert monsters in zoo
	    list.reset();
	    while(list.remain() > 0)
	    {
		if(list.current()->IsMonster()) // Monsters in a queue
		{
		    MONSTER* monster = (MONSTER*) list.current();
		    AddMonster(monster);
		    monster->NewLevel(depth(), SYMBOL_ERROR);
		}
		if(list.current()->IsPlayer()) // Sleeping players
		    bug.error("LEVEL::LoadText", "player");

		list.next();
	    }
	    break;
	}
	case ORDER_EXPERIENCE:
	    parse->skip(INPUT_SPACE);
	    MyExperience = parse->GetLong();
	    break;
	case ORDER_SAVE:
	    parse->skip(INPUT_SPACE);
	    SaveChance = parse->GetLong();
	    break;
	case ORDER_CREATE:
	{
	    parse->skip(INPUT_SPACE);
	    level_type StairUp = global.CaveToLevel(parse->GetWord());
	    if(StairUp == LEVEL_ERROR)
		parse->wrong("LEVEL::LoadText", "StairUp");
	    parse->skip(INPUT_SPACE);
	    level_type StairDown = global.CaveToLevel(parse->GetWord());
	    if(StairUp == LEVEL_ERROR)
		parse->wrong("LEVEL::LoadText", "StairDown");
	    CreateLevel(StairUp, StairDown);
	    break;
	}
	case ORDER_POPULATION:
	    parse->skip(INPUT_SPACE);
	    frequence.load(parse);
	    break;
	case ORDER_END:
	case ORDER_ERROR:
	default:
	    parse->wrong("LEVEL::LoadText",
			 format("Unknown order %s ", order));
	}
	if(parse->LookAhead(INPUT_COMMENT))
	    parse->skip(INPUT_COMMENT);
    }
}    

// @@@ void LEVEL::CreateLevel(level_type StairUp, level_type StairDown)

void LEVEL::CreateLevel(level_type StairUp, level_type StairDown)
{
    int NumberOfCaves = rnd(2, 25);
    int DownStairCave = rnd(1, NumberOfCaves - 1);
    
    coor old = Origo;
    
    location['<'] = Origo;
    CreateCave(Origo, FLOOR, 3);
    
    BUNCH* bunch = create(Origo);
    bunch->add(new STAIR_UP(StairUp));
    
    while(NumberOfCaves-- > 0)
    {
	coor pos(rnd(rnd(75)) - rnd(rnd(75)),
		 rnd(rnd(75)) - rnd(rnd(75)));
	
	if(DownStairCave-- == 1)
	{
	    CreateCave(pos, FLOOR, 1);
	    location['>'] = pos;
	    bunch = create(pos);
	    bunch->add(new STAIR_DOWN(StairDown));
	}
	else
	    CreateCave(pos, FLOOR, 2);
	
	CreatePath(pos, old);
	
	if(rnd.chance(2))
	    old = pos;
    }
    CreatePath(old, Origo);
}

// @@@ void LEVEL::CreateCave(coor where, rock_type value,

void LEVEL::CreateCave(coor where, rock_type value,
			      int width, int chance = 1)
{
    // Dig here unless it's already done
    if(rock(where) > value)
    {
	dig(where, value);
	if(IsRock(value))
	    /* Do Nothing */;
	else if(frequence.MonsterChance())
	{
	    MONSTER* monster = frequence.CreateMonster(experience());
	    
	    AddMonster(monster);
	    monster->NewLevel(depth(), SYMBOL_ERROR);
	    monster->move(this, where);
	}
	else if(frequence.ObjectChance())
	{
	    BUNCH* bunch = create(where);
	    bunch->add(CreateThing(experience()));
	}
    }
    else
	return;

    // Stop after some time
    if(rnd.chance(chance, chance + width + 4))
	return;
    // Or use a weaker pick
    else if(rnd.chance(chance, chance + width + 4))
    {
	value++;
	if(value > SOFT_ROCK)
	    return;
    }
    // Send the happy message to your environment
    for(region reg(coor(-1, -1), coor(2, 2)); reg.check(); reg.NextX())
    {
	if(reg == Origo)
	    continue;
	else if(rnd.chance(chance, chance + width))
	    continue;
	else
	    CreateCave(where + reg, value, width, chance + 1);
    }
}

// @@@ void LEVEL::CreatePath(coor from, coor to)

void LEVEL::CreatePath(coor from, coor to)
{
    coor diff = to - from;
    int delta = (abs(diff.x) + abs(diff.y)) / 5;

    if(delta == 0)
    {
	straight line(from, to);

	for(;;)
	{
	    if(rock(line) > UNEVEN_FLOOR)
		dig(line, UNEVEN_FLOOR);

	    if(!line.check())
		break;
	    else
		line.next();
	}
    }
    else
    {
	// Calculate centrum of the path
	coor pos = from + ((to - from) / 2);
	// Move it a bit...
	pos += coor(rnd(-delta, delta), rnd(-delta, delta));
	// Draw the two part of the new path
	CreatePath(from, pos);
	CreatePath(pos, to);
    }
}

// @@@ void LEVEL::save(BINARY* binary)

void LEVEL::save(BINARY* binary)
{
    // Special locations
    for(int i = 1; i < MaxChar; i++)
    {
	if(location[i] != NoWhere)
	{
	    binary->PutChar(i);
	    binary->PutCoor(location[i]);
	}
    }
    binary->PutChar(0);		// End of locations

    // Save a map of the rock and light
    map.SaveDefault(binary);
    map.save(binary);

    // Save all objects on the level
    for(region reg(Origo, coor(BucketSize, BucketSize));
	reg.check(); reg.NextX())
    {
	bucket[reg.x][reg.y].save(binary);
    }
    binary->PutCoor(NoWhere);	// End of objects

    // Save special events...
    event.save(binary);

    // How the rock look?
    for(rock_type rock = 0; rock < ROCK_END; rock++)
	binary->PutChar(RockSymbol[rock]);

    // Depth, experience and save chance
    binary->PutShort(MyDepth);
    binary->PutShort(MyExperience);
    binary->PutShort(SaveChance);

    // Save monster frequence
    frequence.save(binary);
}

// @@@ void LEVEL::load(BINARY* binary)

void LEVEL::load(BINARY* binary)
{
    // Special locations
    for(;;)
    {
	int i = binary->GetChar();

	if(i == 0)
	    break;
	else
	    location[i] = binary->GetCoor();
    }

    // Load a map of the rock and light
    map.LoadDefault(binary);
    map.load(binary);
    
    // Load all objects from the level
    for(;;)
    {
	coor pos = binary->GetCoor();

	if(pos == NoWhere)
	    break;
	else
	{
	    OBJECT_LIST list;
	    list.load(binary);
	    
	    // Put 'em on the floor
	    BUNCH* bunch = create(pos);
	    bunch->add(&list);

	    // Insert monsters in zoo
	    list.reset();
	    while(list.remain() > 0)
	    {
		if(list.current()->IsMonster()) // Monsters in a queue
		    AddMonster((MONSTER*) list.current());
		if(list.current()->IsPlayer()) // Sleeping players
		   sleepers.add((PLAYER*) list.current());
		list.next();
	    }
	}
    }
    // Restore special events...
    event.load(binary);

    // How the rock look?
    for(rock_type rock = 0; rock < ROCK_END; rock++)
	RockSymbol[rock] = binary->GetChar();

    // Depth, experience and save chance
    MyDepth = binary->GetShort();
    MyExperience = binary->GetShort();
    SaveChance = binary->GetShort();

    // Load frequence of monsters
    frequence.load(binary);

    // Something might have happened while saved
    event.update(this);
}

// @@@ void LEVEL::CheckOut(MONSTER_LIST* export,

void LEVEL::CheckOut(MONSTER_LIST* export,
		     WAIT* wait, SLEEP* sleep, HI_SCORE* HiScore)
{
    // Remove old binary
    (void) unlink(global.CaveBinaryFile(depth()).string());

    // Save the binary version, if appropiate
    if(rnd.chance(SaveChance, 1000))
    {
	party.reset();
	while(party.current() != NULL)
	{
	    party.current()->message("The world fades away...");
	    party.current()->sleep(this, 0);
	    party.next();
	}
	CleanUpPlayerQueue(wait, sleep, HiScore);
	CleanUpMonsterQueue(export);
	
	BINARY binary(BUFFER_WRITE, global.CaveBinaryFile(depth()).string());
	save(&binary);
    }
    else
    {
	zoo.reset();
	while(zoo.current() != NULL)
	{
	    zoo.current()->message("Suddenly, the dungeon collapse...");
	    zoo.current()->kill("was trapped in an earthquake", this);
	    zoo.next();
	}
	CleanUpPlayerQueue(wait, sleep, HiScore);
	CleanUpMonsterQueue(export);
    }
}    

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
