// @ myth - server - command.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "server/question.h"
#include "common/error.h"
#include "server/level.h"
#include "server/player.h"
#include "common/page.h"
#include "server/sleep.h"
#include "server/specify.h"
#include "server/rnd.h"
#include "common/format.h"

// @@ COMMAND

// @@ class COMMAND

// @@@ question_state COMMAND::parse(CHANNEL*, question_state)

question_state COMMAND::parse(CHANNEL*, question_state)
{
    bug.error("COMMAND::parse", "virtual");
    return Q_STATE_DONE;
}

// @@@ cont COMMAND::play(LEVEL*, PLAYER*, long&)

cont COMMAND::play(LEVEL*, PLAYER*, long&)
{
    bug.error("COMMAND::play", "virtual");
    return DONE;
}

// @@ class COMMAND_DIRECTION

// @@@ question_state COMMAND_DIRECTION::parse(CHANNEL* channel,

question_state COMMAND_DIRECTION::parse(CHANNEL* channel, question_state state)
{
    if(state == Q_STATE_START)
	dir = channel->GetChar();
    return Q_STATE_DONE;
}

// @@ class COMMAND_OBJECT

// @@@ question_state COMMAND_OBJECT::parse(CHANNEL* channel,

question_state COMMAND_OBJECT::parse(CHANNEL* channel, question_state state)
{
    // Waiting for a letter
    static const question_state Q_STATE_LETTER = 1;

    if(state == Q_STATE_START)
    {
	symbol = channel->GetChar();

	if(IsKind(symbol))
	    return Q_STATE_LETTER;
    }
    else if(state == Q_STATE_LETTER)
	letter = channel->GetChar();
    else
	bug.error("COMMAND_OBJECT::parse", format("Unknown state %d", state));

    return Q_STATE_DONE;
}

// @@ class COMMAND_MOVE

// @@@ cont COMMAND_MOVE::play(LEVEL* level, PLAYER* player, long& argument)

cont COMMAND_MOVE::play(LEVEL* level, PLAYER* player, long& argument)
{
    if(!player->CanMove())
    {
	player->message("You are unable to move.");
	return DONE;
    }
    coor position = player->position();
    coor target = position;
    BUNCH* bunch;
    if(player->CheckInterrupt() && argument != 0)
	/* Do Nothing */;
    else if(!ValidDirection(dir))
	// The client has sent an invalid direction (error)
	bug.warning("COMMAND_MOVE::play",
		    format("Unknown direction %d", dir));
    else if(dir == DIRECTION_UP)
    {
	bunch = level->create(position);
	bunch->MoveUp(level, player);
    }
    else if(dir == DIRECTION_DOWN)
    {
	bunch = level->create(position);
	bunch->MoveDown(level, player);
    }
    else
    {
	if(player->IsConfused())
	    dir = ConfuseDirection(dir);
	
	target.move(dir);

	bunch = level->create(target);
	rock_type rock = level->rock(target);

	if(IsRock(rock) || !(bunch->AcceptEnter(player)))
	{
	    // If he can not see it, at least he can feel it!
	    player->plot(level, target);
	    bunch->describe(rock, player);
	    level->update(target);
	}
	else
	{
	    bunch = level->create(position);
	    if(bunch->AcceptLeave(player))
	    {
		player->move(level, target);
		level->update(position);
		if(--argument > 0)
		    return CONTINUE;
	    }
	}
    }
    return DONE;
}

// @@ class COMMAND_RUN

// @@@ cont COMMAND_RUN::play(LEVEL* level, PLAYER* player, long& argument)

cont COMMAND_RUN::play(LEVEL* level, PLAYER* player, long& argument)
{
    // Some arguments deafault to 1, other to 100.
    static const int DefaultRunArgument = 100;

    coor target = player->position();
    
    switch(state)
    {
    case RUN_START:
	if(!ValidDirection(dir)) 
	{
	    // The client has sent an invalid direction (error)
	    bug.warning("COMMAND_RUN::play",
			 format("Unknown direction %d", dir));
	    return DONE;
	}
	if(player->IsConfused())
	    dir = ConfuseDirection(dir);

	target.move(dir);

	if(IsRock(level->rock(target)))
	{
	    if(!player->MayFight())
	    {
		player->SetWaiting();
		return CONTINUE;
	    }
	    state = RUN_DIG;
	    level->dig(player, target);
	    if(argument == 0)
		return DONE;
	}
	else
	{
	    BUNCH* from = level->create(player->position());
	    BUNCH* to = level->create(target);

	    if(to->AcceptEnter(player))
	    {
		if(from->AcceptLeave(player))
		{
		    if(!player->CanMove())
		    {
			player->message("You unable to move.");
			return DONE;
		    }
		    player->move(level, target);
		    state = RUN_AUTO;
		}
		else
		{
		    player->UseMovePoint(50);
		    level->update(target);
		    state = RUN_ESCAPE;
		}
	    }
	    else
	    {
		if(!player->MayFight())
		{
		    player->SetWaiting();
		    return CONTINUE;
		}
		QUESTION* question = to->GetQuestion(player, level, target);

		if(question == NULL)
		{
		    if(!player->CanFight())
		    {
			player->message("You unable to fight.");
			return DONE;
		    }
		    state = RUN_FIGHT;
		    to->defend(player, level);
		    if(argument == 0)
			return DONE;
		}
		else
		{
		    FollowUp(question);
		    return DONE;
		}

	    }
	}
	if(argument == 0)
	    argument = DefaultRunArgument;

	break;
    case RUN_AUTO:
    {
	BUNCH* from;
	BUNCH* to;
	
	if(player->IsConfused())
	    dir = ConfuseDirection(dir);

	target.move(dir);

	from = level->create(player->position());
	to = level->create(target);
	
	if(player->CheckInterrupt() ||
	   IsRock(level->rock(target)) ||
	   to->AcceptEnter(player) == FALSE ||
	   from->AcceptLeave(player) == FALSE)
	{
	    level->update(target);
	    return DONE;
	}
	else if(!player->CanMove())
	{
	    player->message("You unable to move.");
	    return DONE;
	}
	else
	    player->move(level, target);

	break;
    }
    case RUN_ESCAPE:
    {
	BUNCH* to = level->create(target);
	
	if(player->CheckInterrupt() || to->AcceptLeave(player))
	    return DONE;
	else
	    player->UseMovePoint(25);
	
	break;
    }
    case RUN_DIG:
	if(player->IsConfused())
	    dir = ConfuseDirection(dir);

	target.move(dir);

	if(player->CheckInterrupt() ||
	   !IsRock(level->rock(target)))
	    return DONE;
	else if(!player->MayFight())
	{
	    player->SetWaiting();
	    return CONTINUE;
	}
	else
	    level->dig(player, target);

	break;
    case RUN_FIGHT:
	if(player->IsConfused())
	    dir = ConfuseDirection(dir);

	target.move(dir);
	BUNCH* enemy = level->create(target);
	
	// Do not check for interrupt, he is fighting to death!

	if(enemy->AcceptEnter(player))
	    return DONE;
	else if(!player->MayFight())
	{
	    player->SetWaiting();
	    return CONTINUE;
	}
	else
	{
	    QUESTION* question = enemy->GetQuestion(player, level, target);
	
	    if(question == NULL)
	    {
		if(!player->CanFight())
		{
		    player->message("You unable to fight.");
		    return DONE;
		}
		enemy->defend(player, level);
	    }
	    else
	    {
		FollowUp(question);
		return DONE;
	    }
	}
	break;
    default:
	player->error("COMMAND_RUN::play", format("state = %d", state));
	return DONE;
    }
    return (--argument < 1) ? DONE : CONTINUE;
}

// @@ class COMMAND_DROP

// @@@ cont COMMAND_DROP::play(LEVEL* level, PLAYER* player, long& argument)

cont COMMAND_DROP::play(LEVEL* level, PLAYER* player, long& argument)
{
    OBJECT_LIST list;
    SPECIFY_LETTER specify(symbol, letter);
    player->pack.remove(&specify, &list, player, argument);
    player->UseMovePoint(5 * list.length());

    if(list.length() < 1)
	player->message("You drop nothing.");
    else 
    {
	PAGE page;
	BUNCH* bunch = level->create(player->position());
	bunch->add(&list, &page, player);

	if(page.length() == 1)
	{
	    page.reset();
	    list.reset();
	    player->message(format("You drop: %s.", page.string()));
	    level->WatchMonster(player, "%s drops %.", list.current());
	}
	else
	{
	    if(player->VerboseGet())
		player->PageMessage("You drop:", "> ", &page);
	    else
		player->message("You drop several items.");
	    
	    level->WatchMonster(player, "%s drops several items.");
	}
    }
    return DONE;
}

// @@ class COMMAND_GET

// @@@ question_state COMMAND_GET::parse(CHANNEL*, question_state),

question_state COMMAND_GET::parse(CHANNEL*, question_state)
{ return Q_STATE_DONE; }

// @@@ cont COMMAND_GET::play(LEVEL* level, PLAYER* player, long& argument)

cont COMMAND_GET::play(LEVEL* level, PLAYER* player, long& argument)
{
    BUNCH* bunch = level->create(player->position());
    bunch->send(INVENTORY_OTHER, player);
    FollowUp(new QUESTION_GET(argument, player->GetChannel()));
    return DONE;
}

// @@ class COMMAND_APPLY

// @@@ cont COMMAND_APPLY::play(LEVEL* level, PLAYER* player, long&)

cont COMMAND_APPLY::play(LEVEL* level, PLAYER* player, long&)
{
    if(!player->MayFight())
    {
	player->SetWaiting();
	return CONTINUE;
    }
    OBJECT_LIST list;
    SPECIFY_LETTER specify(symbol, letter);
    player->pack.RemoveHalf(&specify, &list, 1L);

    if(list.length() < 1)
	player->message("You do nothing.");
    else if(list.length() > 1)
	player->message("You can't apply multiple items at the same time!");
    else
    {
	list.reset();
	if(list.current()->count() > 1L)
	    // Only apply one item at a time
	    FollowUp(list.current()->split(1L)->DoApply(level, player,
						       player->GetChannel()));
	else
	{
	    FollowUp(list.current()->DoApply(level, player,
					     player->GetChannel()));
	    list.remove();
	}
	// It's the responsibility of "apply" to put it back again.
    }
    player->pack.add(&list);	// Other items to the pack

    return DONE;
}

// @@ class COMMAND_THROW

// @@@ question_state COMMAND_THROW::parse(CHANNEL* channel,

question_state COMMAND_THROW::parse(CHANNEL* channel, question_state state)
{
    // Additionel states...
    static const int Q_STATE_LETTER = 1;
    static const int Q_STATE_POSITION = 2;
    
    switch(state)
    {
    case Q_STATE_START:
	symbol = channel->GetChar();

	if(IsKind(symbol))
	    return Q_STATE_LETTER;
	else
	    return Q_STATE_POSITION;
    case Q_STATE_LETTER:
	letter = channel->GetChar();
	return Q_STATE_POSITION;
    case Q_STATE_POSITION:
	if(channel->IsCoorReady())
	{
	    position = channel->GetCoor();
	    return Q_STATE_DONE;
	}
	else
	    return Q_STATE_POSITION;
    default:
	bug.error("COMMAND_THROW::parse", format("Unknown state %d", state));
	return Q_STATE_DONE;
    }
}

// @@@ cont COMMAND_THROW::play(LEVEL* level, PLAYER* player, long&)

cont COMMAND_THROW::play(LEVEL* level, PLAYER* player, long&)
{
    if(!player->MayFight())
    {
	player->SetWaiting();
	return CONTINUE;
    }
    OBJECT_LIST list;
    SPECIFY_LETTER specify(symbol, letter);
    player->pack.remove(&specify, &list, player, 1L);

    if(list.length() < 1)
	player->message("You throw nothing.");
    else if(list.length() > 1)
	player->message("You can't figure out how to throw all that!");
    else
    {
	list.reset();
	if(list.current()->count() > 1L)
	    // Only throw one item
	    level->throw(player, list.current()->split(1L), position);
	else
	{
	    level->throw(player, list.current(), position);
	    list.remove();
	}
    }
    player->pack.add(&list);	// Return surviving items to the pack

    return DONE;
}

// @@ class COMMAND_SLEEP

// @@@ question_state COMMAND_SLEEP::parse(CHANNEL* channel,

question_state COMMAND_SLEEP::parse(CHANNEL* channel, question_state state)
{
    if(state == Q_STATE_START)
	conf = channel->GetChar();
    else
	bug.warning("COMMAND_SLEEP::parse",
		    format("Got %d", channel->GetChar()));

    return Q_STATE_DONE;
}

// @@@ cont COMMAND_SLEEP::play(LEVEL* level, PLAYER* player, long&)

cont COMMAND_SLEEP::play(LEVEL* level, PLAYER* player, long&)
{
    switch(conf)
    {
    case YES:
	FollowUp(new QUESTION_PASSWORD(player->GetChannel()));
	break;
    case NO:
	player->sleep(level, 0);
	break;
    case QUIT:
	break;
    default:
	bug.warning("COMMAND_SLEEP::player", format("conf = %d", conf));
    }
    return DONE;
}

// @@ class COMMAND_COMMAND

// @@@ question_state COMMAND_COMMAND::parse(CHANNEL* channel,

question_state COMMAND_COMMAND::parse(CHANNEL* channel, question_state state)
{
    // Use state as an index to the string
    // This require that Q_STATE_START == 0!

    if(state == Q_STATE_DONE || (string[state] = channel->GetChar()) == 0)
	return Q_STATE_DONE;
    else if(state < MessageSize - 1)
	return state + 1;
    else
	return MessageSize - 1;
}

// @@@ cont COMMAND_COMMAND::play(LEVEL* level, PLAYER* player,

cont COMMAND_COMMAND::play(LEVEL* level, PLAYER* player, long&)
{
    FollowUp(player->command(level, string));
    return DONE;
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
