// @ myth - server - question.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "server/question.h"
#include "common/error.h"
#include "server/level.h"
#include "server/player.h"
#include "common/page.h"
#include "server/sleep.h"
#include "server/specify.h"
#include "server/ray.h"
#include "server/rnd.h"
#include "common/format.h"

// @@ class QUESTION

// @@@ bool QUESTION::fresh()

bool QUESTION::fresh()
{ return state == Q_STATE_ANSWER; }

// @@@ void QUESTION::parse(CHANNEL* channel)

void QUESTION::parse(CHANNEL* channel)
{
    (void) channel->GetChar();
    bug.error("QUESTION::parse", "virtual");
}

// @@@ question_value QUESTION::play(LEVEL* level, PLAYER* player)

question_value QUESTION::play(LEVEL*, PLAYER*)
{
    bug.error("QUESTION::play", "virtual");
    return Q_VALUE_DONE;
}

// @@@ QUESTION* QUESTION::next()

QUESTION* QUESTION::next()
{ return NULL; /* Default: No further questions. */ }

// @@@ QUESTION::QUESTION()

QUESTION::QUESTION()
{
    state = Q_STATE_ANSWER;
    argument = 0L;
}

// @@@ QUESTION::~QUESTION()

QUESTION::~QUESTION()
{ }

// @@ class QUESTION_OBJECT

// @@@ void QUESTION_OBJECT::parse(CHANNEL* channel)

void QUESTION_OBJECT::parse(CHANNEL* channel)
{
    // Waiting for a letter
    static const int Q_STATE_LETTER = 1;

    switch(state)
    {
    case Q_STATE_ESCAPE:
    case Q_STATE_DONE:
	return;
    case Q_STATE_ANSWER:
	switch(channel->GetChar())
	{
	case ANSWER_ESCAPE:
	    state = Q_STATE_ESCAPE;
	    break;
	case ANSWER:
	    state = Q_STATE_START;
	    break;
	case ANSWER_ARGUMENT:
	    state = Q_STATE_ARGUMENT;
	    break;
	default:
	    bug.warning("QUESTION_OBJECT::parse", "Unknown ANSWER");
	}
	return;
    case Q_STATE_ARGUMENT:
	if(channel->IsLongReady())
	{
	    long arg = channel->GetLong();
	    if(arg != 0L)
		argument = arg;
	    state = Q_STATE_START;
	}
	return;
    case Q_STATE_START:
	symbol = channel->GetChar();
	state = IsKind(symbol) ? Q_STATE_LETTER : Q_STATE_DONE;
	return;
    case Q_STATE_LETTER:
	letter = channel->GetChar();
	state = Q_STATE_DONE;
	return;
    default:
	bug.error("QUESTION_OBJECT::parse", format("Unknown state %d", state));
    }
}

// @@@ QUESTION_OBJECT::QUESTION_OBJECT()

QUESTION_OBJECT::QUESTION_OBJECT() : ()
{ }

// @@ class QUESTION_LETTER

// @@@ void QUESTION_LETTER::parse(CHANNEL* channel)

void QUESTION_LETTER::parse(CHANNEL* channel)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
    case Q_STATE_DONE:
	return;
    case Q_STATE_ANSWER:
	switch(channel->GetChar())
	{
	case ANSWER_ESCAPE:
	    state = Q_STATE_ESCAPE;
	    break;
	case ANSWER:
	    state = Q_STATE_START;
	    break;
	case ANSWER_ARGUMENT:
	    state = Q_STATE_ARGUMENT;
	    break;
	default:
	    bug.warning("QUESTION_LETTER::parse", "Unknown ANSWER");
	}
	return;
    case Q_STATE_ARGUMENT:
	if(channel->IsLongReady())
	{
	    long arg = channel->GetLong();
	    if(arg != 0L)
		argument = arg;
	    state = Q_STATE_START;
	}
	return;
    case Q_STATE_START:
	letter = channel->GetChar();
	state = Q_STATE_DONE;
	return;
    case Q_STATE_COMMAND:
    default:
	bug.error("QUESTION_LETTER::parse", format("Unknown state %d", state));
    }
}

// @@@ QUESTION_LETTER::QUESTION_LETTER()

QUESTION_LETTER::QUESTION_LETTER() : ()
{ }

// @@ class QUESTION_POSITION

// @@@ void QUESTION_POSITION::parse(CHANNEL* channel)

void QUESTION_POSITION::parse(CHANNEL* channel)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
    case Q_STATE_DONE:
	return;
    case Q_STATE_ANSWER:
	switch(channel->GetChar())
	{
	case ANSWER_ESCAPE:
	    state = Q_STATE_ESCAPE;
	    break;
	case ANSWER:
	    state = Q_STATE_START;
	    break;
	case ANSWER_ARGUMENT:
	    state = Q_STATE_ARGUMENT;
	    break;
	default:
	    bug.warning("QUESTION_POSITION::parse", "Unknown ANSWER");
	}
	return;
    case Q_STATE_ARGUMENT:
	if(channel->IsLongReady())
	{
	    long arg = channel->GetLong();
	    if(arg != 0L)
		argument = arg;
	    state = Q_STATE_START;
	}
	return;
    case Q_STATE_START:
	if(channel->IsCoorReady())
	{
	    position = channel->GetCoor();
	    state = Q_STATE_DONE;
	}
	return;
    case Q_STATE_COMMAND:
    default:
	bug.error("QUESTION_POSITION::parse",
		  format("Unknown state %d", state));
    }
}

// @@@ QUESTION_POSITION::QUESTION_POSITION()

QUESTION_POSITION::QUESTION_POSITION() : ()
{ }

// @@ class QUESTION_COMMAND

// @@@ void QUESTION_COMMAND::parse(CHANNEL* channel)

void QUESTION_COMMAND::parse(CHANNEL* channel)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
    case Q_STATE_DONE:
	return;
    case Q_STATE_ANSWER:
	switch(channel->GetChar())
	{
	case ANSWER_ESCAPE:
	    state = Q_STATE_ESCAPE;
	    break;
	case ANSWER:
	    state = Q_STATE_COMMAND;
	    break;
	case ANSWER_ARGUMENT:
	    state = Q_STATE_ARGUMENT;
	    break;
	default:
	    bug.warning("QUESTION_COMMAND::parse", "Unknown ANSWER");
	}
	return;
    case Q_STATE_ARGUMENT:
	if(channel->IsLongReady())
	{
	    argument = channel->GetLong();
	    state = Q_STATE_COMMAND;
	}
	return;
    case Q_STATE_COMMAND:
	switch(channel->GetChar())
	{
	case CLIENT_MOVE:
	    command = new COMMAND_MOVE;
	    break;
	case CLIENT_RUN:
	    command = new COMMAND_RUN;
	    break;
	case CLIENT_DROP:
	    command = new COMMAND_DROP;
	    break;
	case CLIENT_GET:
	    command = new COMMAND_GET;
	    state = Q_STATE_DONE;
	    return;
	case CLIENT_APPLY:
	    command = new COMMAND_APPLY;
	    break;
	case CLIENT_THROW:
	    command = new COMMAND_THROW;
	    break;
	case CLIENT_SLEEP:
	    command = new COMMAND_SLEEP;
	    break;
	case CLIENT_COMMAND:
	    command = new COMMAND_COMMAND;
	    break;
	default:
	    bug.warning("QUESTION_COMMAND::parse", "Unknown command");
	    state = Q_STATE_DONE;
	    return;
	}
	state = Q_STATE_START;
	return;
    case Q_STATE_START:
    default:
	if(command == NULL)
	{
	    bug.error("QUESTION_COMMAND::parse", "NULL command");
	    state = Q_STATE_DONE;
	}
	else
	    state = command->parse(channel, state);

	return;
    }
}	

// @@@ question_value QUESTION_COMMAND::play(LEVEL* level, PLAYER* player)

question_value QUESTION_COMMAND::play(LEVEL* level, PLAYER* player)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
	return Q_VALUE_DONE;
    case Q_STATE_DONE:
	if(command == NULL)
	{
	    bug.error("QUESTION_COMMAND::play", "NULL command");
	    return Q_VALUE_DONE;
	}
	else if(command->play(level, player, argument) == DONE)
	    return Q_VALUE_DONE;
	else
	    return Q_VALUE_AGAIN;
    case Q_STATE_ANSWER:
    case Q_STATE_ARGUMENT:
    case Q_STATE_COMMAND:
    case Q_STATE_START:
    default:
	return Q_VALUE_PARSING;
    }
}

// @@@ QUESTION* QUESTION_COMMAND::next()

QUESTION* QUESTION_COMMAND::next()
{ return command == NULL ? NULL : command->next(); }

// @@@ QUESTION_COMMAND::QUESTION_COMMAND(CHANNEL* channel) 

QUESTION_COMMAND::QUESTION_COMMAND(CHANNEL* channel) 
{
    channel->PutChar(SERVER_REQUEST);
    channel->PutChar(PROMPT_COPYRIGHT);
    channel->PutChar(REQUEST_COMMAND);
    command = NULL;
}

// @@@ QUESTION_COMMAND::~QUESTION_COMMAND()

QUESTION_COMMAND::~QUESTION_COMMAND()
{
    if(command)			// There is no reason for this test!
	delete command;
}

// @@ class QUESTION_ZAP

// @@@ question_value QUESTION_ZAP::play(LEVEL* level, PLAYER* player)

question_value QUESTION_ZAP::play(LEVEL* level, PLAYER* player)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
	level->zap(player, ray, Origo);
	ray = NULL;
	return Q_VALUE_DONE;
    case Q_STATE_DONE:
    { 
	level->zap(player, ray, position);
	ray = NULL;
	return Q_VALUE_DONE;
    }
    case Q_STATE_ANSWER:
    case Q_STATE_ARGUMENT:
    case Q_STATE_COMMAND:
    case Q_STATE_START:
    default:
	return Q_VALUE_PARSING;
    }
}

// @@@ QUESTION_ZAP::QUESTION_ZAP(long arg, CHANNEL* channel)

QUESTION_ZAP::QUESTION_ZAP(RAY* r, CHANNEL* channel) : ()
{
    ray = r;
    channel->PutChar(SERVER_REQUEST);
    channel->PutChar(PROMPT_ZAP);
    channel->PutChar(REQUEST_POSITION);
}

// @@@ QUESTION_ZAP::~QUESTION_ZAP()

QUESTION_ZAP::~QUESTION_ZAP()
{
    if(ray != NULL)
	delete ray;
}

// @@ class QUESTION_EAT

// @@@ question_value QUESTION_EAT::play(LEVEL* level, PLAYER* player)

question_value QUESTION_EAT::play(LEVEL* level, PLAYER* player)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
	return Q_VALUE_DONE;
    case Q_STATE_DONE:
    {
	OBJECT_LIST list;
	SPECIFY_LETTER specify(SYMBOL_FOOD, letter);
	BUNCH* bunch = level->create(player->position());
	bunch->remove(&specify, &list, NULL, 1L);
	
	if(list.length() < 1)
	    player->message("You don't eat anything.");
	else if(list.length() > 1)
	{
	    player->message("Don't stuff yourself!");
	}
	else
	{
	    list.reset();
	    OBJECT* object = list.current();

	    if(list.current()->count() > 1L)
		// Only apply one item at a time
		TryAgain = object->split(1L)->DoApply(level, player,
						      player->GetChannel());
	    else
	    {
		TryAgain = object->DoApply(level, player,
					   player->GetChannel());
		list.remove();
	    }
	}
	// Put the rest back.
	bunch = level->create(player->position());
	bunch->add(&list);
	return Q_VALUE_DONE;
    }
    case Q_STATE_ANSWER:
    case Q_STATE_ARGUMENT:
    case Q_STATE_COMMAND:
    case Q_STATE_START:
    default:
	return Q_VALUE_PARSING;
    }
}

// @@@ QUESTION* QUESTION_EAT::next()

QUESTION* QUESTION_EAT::next()
{ return TryAgain; }

// @@@ QUESTION_EAT::QUESTION_EAT(CHANNEL* channel)

QUESTION_EAT::QUESTION_EAT(CHANNEL* channel)
{
    TryAgain = NULL;
    channel->PutChar(SERVER_REQUEST);
    channel->PutChar(PROMPT_EAT);
    channel->PutChar(REQUEST_LETTER);
    channel->PutChar(INVENTORY_OTHER);
    channel->PutChar(SYMBOL_FOOD);
}

// @@ class QUESTION_IDENTIFY

// @@@ question_value QUESTION_IDENTIFY::play(LEVEL*, PLAYER* player)

question_value QUESTION_IDENTIFY::play(LEVEL*, PLAYER* player)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
	player->message("You don't identify anything.");
	return Q_VALUE_DONE;
    case Q_STATE_DONE:
    {
	OBJECT_LIST list;
	SPECIFY_LETTER specify(symbol, letter);
	player->pack.RemoveHalf(&specify, &list, argument);

	if(list.length() < 1)
	    player->message("You identify nothing.");
	else if(list.length() == 1)
	{
	    list.reset();
	    player->identify(list.current());
	    player->learn(list.current(), value);
	    player->message(format("You identify %s.",
				   print(list.current(),
					 PRINT_OBJECT, player)));
	}
	else
	{
	    player->message(STRING("You are getting confused...",
				   STRING_STATIC));
	    player->AddAbility(new COUNTER_CONFUSED(rnd(250, 500)));
	}
	player->pack.add(&list);
	return Q_VALUE_DONE;
    }
    case Q_STATE_ANSWER:
    case Q_STATE_ARGUMENT:
    case Q_STATE_COMMAND:
    case Q_STATE_START:
    default:
	return Q_VALUE_PARSING;
    }
}

// @@@ QUESTION_IDENTIFY::QUESTION_IDENTIFY(int val, CHANNEL* channel)

QUESTION_IDENTIFY::QUESTION_IDENTIFY(int val, CHANNEL* channel)
{
    value = val;
    channel->PutChar(SERVER_REQUEST);
    channel->PutChar(PROMPT_IDENTIFY);
    channel->PutChar(REQUEST_OBJECT);
    channel->PutChar(INVENTORY_OWN);
}

// @@ class QUESTION_BLESS

// @@@ void QUESTION_BLESS::CurseWorn(PLAYER* player)

void QUESTION_BLESS::CurseWorn(PLAYER* player)
{
    player->message("You feel the anger of the gods!");

    SPECIFY_WORN specify(player);
    OBJECT_LIST list;
    
    player->pack.RemoveHalf(&specify, &list);
    
    list.reset();
    while(list.current() != NULL)
    {
	list.current()->curse();
	next();
    }
    player->pack.add(&list);
}
    

// @@@ question_value QUESTION_BLESS::play(LEVEL*, PLAYER* player)

question_value QUESTION_BLESS::play(LEVEL*, PLAYER* player)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
	CurseWorn(player);
	return Q_VALUE_DONE;
    case Q_STATE_DONE:
    {
	OBJECT_LIST list;
	SPECIFY_LETTER specify(symbol, letter);
	player->pack.RemoveHalf(&specify, &list, argument);

	if(list.length() < 1)
	    CurseWorn(player);
	else if(list.length() == 1)
	{
	    list.reset();
	    if(!player->CheckAbility(ABILITY_BLIND))
	    {
		player->message(format("Your %s glows %s for a moment.",
				       print(list.current(),
					     PRINT_VALUE | PRINT_TYPE, player),
				       (holy == CURSED) ? "red" : "blue"));
	    }
	    player->identify(list.current());
	    switch(holy)
	    {
	    case CURSED:
		list.current()->curse();
		break;
	    case NORMAL:
		list.current()->RemoveCurse();
		break;
	    case BLESSED:
		list.current()->bless();
		break;
	    default:
		bug.error("QUESTION_BLESS::play", "Impossible holy state");
	    }
	    player->pack.add(&list);
	}
	else
	{
	    player->pack.add(&list);
	    CurseWorn(player);
	}
	return Q_VALUE_DONE;
    }
    case Q_STATE_ANSWER:
    case Q_STATE_ARGUMENT:
    case Q_STATE_COMMAND:
    case Q_STATE_START:
    default:
	return Q_VALUE_PARSING;
    }
}

// @@@ QUESTION_BLESS::QUESTION_BLESS(holy_type h, CHANNEL* channel)

QUESTION_BLESS::QUESTION_BLESS(holy_type h, CHANNEL* channel)
{
    holy = h;
    channel->PutChar(SERVER_REQUEST);
    channel->PutChar(PROMPT_BLESS);
    channel->PutChar(REQUEST_OBJECT);
    channel->PutChar(INVENTORY_OWN);
}

// @@ class QUESTION_ENCHANT

// @@@ question_value QUESTION_ENCHANT::play(LEVEL*, PLAYER* player)

question_value QUESTION_ENCHANT::play(LEVEL*, PLAYER* player)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
	player->message("You stop the magic.");
	return Q_VALUE_DONE;
    case Q_STATE_DONE:
    {
	OBJECT_LIST list;
	SPECIFY_LETTER specify(symbol, letter);
	player->pack.RemoveHalf(&specify, &list, argument);

	if(list.length() == 1)
	{
	    list.reset();
	    list.current()->ChangeValue(value);
	    if(!player->CheckAbility(ABILITY_BLIND))
	    {
		player->message(format("Your %s glows %s for a moment.",
				       print(list.current(),
					     PRINT_VALUE | PRINT_TYPE, player),
				       (value < 0) ? "yellow" : "green"));
	    }
	}
	else
	    player->message("The magic fails...");

	player->pack.add(&list);
	return Q_VALUE_DONE;
    }
    case Q_STATE_ANSWER:
    case Q_STATE_ARGUMENT:
    case Q_STATE_COMMAND:
    case Q_STATE_START:
    default:
	return Q_VALUE_PARSING;
    }
}

// @@@ QUESTION_ENCHANT::QUESTION_ENCHANT(int val, CHANNEL* channel)

QUESTION_ENCHANT::QUESTION_ENCHANT(int val, CHANNEL* channel)
{
    value = val;
    channel->PutChar(SERVER_REQUEST);
    channel->PutChar(PROMPT_MAGIC);
    channel->PutChar(REQUEST_OBJECT);
    channel->PutChar(INVENTORY_OWN);
}

// @@ class QUESTION_TELEPORT

// @@@ question_value QUESTION_TELEPORT::play(LEVEL* level, PLAYER* player)

question_value QUESTION_TELEPORT::play(LEVEL* level, PLAYER* player)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
	return Q_VALUE_DONE;
    case Q_STATE_DONE:
	if(sqr(position.x) + sqr(position.y) > sqr(value))
	{
	    player->message("Sorry...");
	    player->teleport(level, value, NORMAL);
	}
	else
	    player->move(level, player->position() + position);

	return Q_VALUE_DONE;
    case Q_STATE_ANSWER:
    case Q_STATE_ARGUMENT:
    case Q_STATE_COMMAND:
    case Q_STATE_START:
    default:
	return Q_VALUE_PARSING;
    }
}

// @@@ QUESTION_TELEPORT::QUESTION_TELEPORT(int val, CHANNEL* channel)

QUESTION_TELEPORT::QUESTION_TELEPORT(int val, CHANNEL* channel)
{
    value = val;
    channel->PutChar(SERVER_REQUEST);
    channel->PutChar(PROMPT_TELEPORT);
    channel->PutChar(REQUEST_POSITION);
}

// @@ class QUESTION_BREATHE

// @@@ question_value QUESTION_BREATHE::play(LEVEL* level, PLAYER* player)

question_value QUESTION_BREATHE::play(LEVEL* level, PLAYER* player)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
	return Q_VALUE_DONE;
    case Q_STATE_DONE:
	player->breathe(level, position);
	return Q_VALUE_DONE;
    case Q_STATE_ANSWER:
    case Q_STATE_ARGUMENT:
    case Q_STATE_COMMAND:
    case Q_STATE_START:
    default:
	return Q_VALUE_PARSING;
    }
}

// @@@ QUESTION_BREATHE::QUESTION_BREATHE(CHANNEL* channel)

QUESTION_BREATHE::QUESTION_BREATHE(CHANNEL* channel)
{
    channel->PutChar(SERVER_REQUEST);
    channel->PutChar(PROMPT_BREATHE);
    channel->PutChar(REQUEST_POSITION);
}

// @@ class QUESTION_GET

// @@@ question_value QUESTION_GET::play(LEVEL* level, PLAYER* player)

question_value QUESTION_GET::play(LEVEL* level, PLAYER* player)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
	return Q_VALUE_DONE;
    case Q_STATE_DONE:
    { 
	if(!player->CanMove())
	{
	    player->message("You can't carry any more!");
	    return Q_VALUE_DONE;
	}

	OBJECT_LIST list;
	SPECIFY_LETTER specify(symbol, letter);
	BUNCH* bunch = level->create(player->position());
	bunch->remove(&specify, &list,
						  NULL, argument);
	player->UseMovePoint(list.length() * 10);
	
	long weight = 0L;
	list.reset();
	while(list.current() != NULL)
	{
	    weight += list.current()->weight();
	    list.next();
	}

	if(!player->CanMove(weight))
	{
	    player->message("You can't carry that.");
	    BUNCH* bunch = level->create(player->position());
	    bunch->add(&list);
	    return Q_VALUE_DONE;
	}
	else if(list.length() < 1)
	    player->message("You don't pick anything up.");
	else 
	{
	    PAGE page;
	    player->pack.add(&list, &page, player);
	    if(page.length() == 1)
	    {
		page.reset();
		list.reset();
		player->message(format("You pick up: %s.", page.string()));
		level->WatchMonster(player, "%s picks %s up.", list.current());
	    }
	    else
	    {
		if(player->VerboseGet())
		    player->PageMessage("You get:", "> ", &page);
		else
		    player->message("You pick up several items.");
		
		level->WatchMonster(player, "%s picks up several items.");
	    }
	}

	return Q_VALUE_DONE;
    }
    case Q_STATE_ANSWER:
    case Q_STATE_ARGUMENT:
    case Q_STATE_COMMAND:
    case Q_STATE_START:
    default:
	return Q_VALUE_PARSING;
    }
}

// @@@ QUESTION_GET::QUESTION_GET(long arg, CHANNEL* channel)

QUESTION_GET::QUESTION_GET(long arg, CHANNEL* channel) : ()
{
    channel->PutChar(SERVER_REQUEST);
    channel->PutChar(PROMPT_GET);
    channel->PutChar(REQUEST_OBJECT);
    channel->PutChar(INVENTORY_OTHER);
    argument = arg;
}

// @@ class QUESTION_STRING

// @@@ void QUESTION_STRING::parse(CHANNEL* channel)

void QUESTION_STRING::parse(CHANNEL* channel)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
    case Q_STATE_DONE:
	return;
    case Q_STATE_ANSWER:
	switch(channel->GetChar())
	{
	case ANSWER_ESCAPE:
	    state = Q_STATE_ESCAPE;
	    break;
	case ANSWER:
	    state = Q_STATE_START;
	    break;
	case ANSWER_ARGUMENT:
	    state = Q_STATE_ARGUMENT;
	    break;
	default:
	    bug.warning("QUESTION_STRING::parse", "Unknown ANSWER");
	}
	return;
    case Q_STATE_ARGUMENT:
	if(channel->IsLongReady())
	{
	    argument = channel->GetLong();
	    state = Q_STATE_START;
	}
	return;
    case Q_STATE_COMMAND:
	bug.warning("QUESTION_STRING::parse", "Q_STATE_COMMAND unused");
	state = Q_STATE_DONE;
	return;
    case Q_STATE_START:
    default:
	// Use state as an index to the string
	// This require that Q_STATE_START == 0!

	// Cut to long messages
	if(state >= MessageSize - 1)
	    state = MessageSize - 1;
	
	if((string[state] = channel->GetChar()) == 0)
	    state = Q_STATE_DONE;
	else
	    state++;
    }
}

// @@@ QUESTION_STRING::QUESTION_STRING(CHANNEL* channel,

QUESTION_STRING::QUESTION_STRING(CHANNEL* channel,
				 prompt_type prompt,
				 request_type request = REQUEST_STRING)
{
    // Initialise
    string[0] = 0;

    // Ask for a string
    channel->PutChar(SERVER_REQUEST);
    channel->PutChar(prompt);
    channel->PutChar(request);
}

// @@ class QUESTION_RESTORE

// @@@ question_value QUESTION_RESTORE::play(LEVEL*, PLAYER* player)

question_value QUESTION_RESTORE::play(LEVEL*, PLAYER* player)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
	TryAgain = new QUESTION_START_UP(player, sleep);
	return Q_VALUE_DONE;
    case Q_STATE_DONE:
    {
	STRING passwd(string);
	id_type id = sleep->identify(name, passwd);

	if(id == NoIdentity)
	{
	    player->message("Wrong password, try again.");
	    TryAgain = new QUESTION_RESTORE(player->GetChannel(),
					    sleep, name());
	}
	else
	{
	    player->restore(sleep->where(id), id);
	    sleep->remove(id);
	}
	return Q_VALUE_DONE;
    }
    case Q_STATE_ANSWER:
    case Q_STATE_ARGUMENT:
    case Q_STATE_COMMAND:
    case Q_STATE_START:
    default:
	return Q_VALUE_PARSING;
    }
}

// @@@ QUESTION* QUESTION_RESTORE::next()

QUESTION* QUESTION_RESTORE::next()
{ return TryAgain; }

// @@@ QUESTION_RESTORE::QUESTION_RESTORE(CHANNEL* channel,

QUESTION_RESTORE::QUESTION_RESTORE(CHANNEL* channel, SLEEP* saved, 
				   const char* who)
		                : (channel, PROMPT_PASSWORD)
{
    TryAgain = NULL;		// Next question
    sleep = saved;		// List of sleepers
    name = STRING(who);		// Name from QUESTION_START_UP
}

// @@ class QUESTION_START_UP

// @@@ question_value QUESTION_START_UP::play(LEVEL*, PLAYER* player)

question_value QUESTION_START_UP::play(LEVEL*, PLAYER* player)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
	TryAgain = new QUESTION_START_UP(player, sleep);
	return Q_VALUE_DONE;
    case Q_STATE_DONE:
    {
	STRING name(string);
	player->ReName(name);
	
	if(sleep->used(name))
	{
	    id_type id = sleep->identify(name, STRING(""));

	    if(id == NoIdentity)
		TryAgain = new QUESTION_RESTORE(player->GetChannel(),
						sleep, name());
	    else
	    {
		player->restore(sleep->where(id), id);
		sleep->remove(id);
	    }
	}
	return Q_VALUE_DONE;
    }
    case Q_STATE_ANSWER:
    case Q_STATE_ARGUMENT:
    case Q_STATE_COMMAND:
    case Q_STATE_START:
    default:
	return Q_VALUE_PARSING;
    }
}

// @@@ QUESTION_START_UP::QUESTION_START_UP(PLAYER* player, SLEEP* saved)

QUESTION_START_UP::QUESTION_START_UP(PLAYER* player, SLEEP* saved)
                                  : (player->GetChannel(), 
				     PROMPT_WELCOME, REQUEST_NAME)
{
    TryAgain = NULL;		// Next question
    sleep = saved;		// Remember the list of sleeping players
    PAGE page;			// Tell which names is used
    sleep->page(&page);
    player->page(&page);
}

// @@ class QUESTION_PASSWORD

// @@@ question_value QUESTION_PASSWORD::play(LEVEL* level, PLAYER* player)

question_value QUESTION_PASSWORD::play(LEVEL* level, PLAYER* player)
{
    switch(state)
    {
    case Q_STATE_ESCAPE:
	return Q_VALUE_DONE;
    case Q_STATE_DONE:
	player->SetPasswd(STRING(string));
	player->sleep(level, 0);
	return Q_VALUE_DONE;
    case Q_STATE_ANSWER:
    case Q_STATE_ARGUMENT:
    case Q_STATE_COMMAND:
    case Q_STATE_START:
    default:
	return Q_VALUE_PARSING;
    }
}

// @@@ QUESTION_PASSWORD::QUESTION_PASSWORD(CHANNEL* channel)

QUESTION_PASSWORD::QUESTION_PASSWORD(CHANNEL* channel) 
			          : (channel, PROMPT_PASSWORD) { }

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
